#!/usr/bin/env python

from __future__ import with_statement

import sys
import re

OTHER_TYPE_STR = "OTHER"

# OBO synonym types that are "likely safe" in the sense that they are
# expected to identify the same concept as apposed to a broader or
# merely related one.
SAFE_SYNONYM_TYPE = set([
        "synonym_exact",
        "synonym_exact_plural",
        "synonym_exact_singular",
        "synonym_plural",
        "synonym_singular",
        ])

def argparser():
    import argparse

    ap=argparse.ArgumentParser(description="Special-purpose evaluation script for \"OBO anatomy\" work. Reads minimal-format SOLS output and analyses the results.")
    ap.add_argument("-p", "--include-part-of", default=False, action="store_true", help="Include part_of predictions.")
    ap.add_argument("-s", "--include-all-synonyms", default=False, action="store_true", help="Include all synonym match-based predictions.")
    ap.add_argument("-t", "--targets", metavar="FILE", default=None, help="File listing \"target\" predictions. If given, misses will be grouped as \"%s\"." % OTHER_TYPE_STR)
    ap.add_argument("-e", "--exclude", metavar="ONTO", help="Exclude responses from ontology ONTO.")
    ap.add_argument("-o", "--only-include", metavar="ONTO", help="Exclude responses from ontologies other than ONTO.")
    ap.add_argument("-f", "--filter", metavar="TERM[,TERM,...]", default=None, help="Filter out given terms.")
    ap.add_argument("-c", "--confusion-matrix", default=False, action="store_true", help="Output tab-separated confusion matrix.")
    ap.add_argument("-T", "--output-types", default=False, action="store_true", help="Output types for each query.")
    ap.add_argument("-E", "--evaluate-against", metavar="FILE", default=None, help="Evaluate precision and recall against labeled cases in file.")
    ap.add_argument("-m", "--matching-only", default=False, action="store_true", help="Only evaluate matching (TP) terms (use with -E)")
    ap.add_argument("-v", "--verbose", default=False, action="store_true", help="Verbose output.")
    ap.add_argument("file", metavar="FILE", nargs=1, help="SOLS output (should be in minimal mode)")
    return ap

class Prediction:
    def __init__(self, type, tid, onto, syn_spec=None, part_of_spec=None):
        self.type = type
        self.tid = tid
        self.onto = onto
        self.syn_spec = syn_spec
        self.part_of_spec = part_of_spec

    def __str__(self):
        synstr = "" if self.syn_spec is None else " [%s]" % self.syn_spec
        postr  = "" if self.part_of_spec is None else "-%s- " % self.part_of_spec
        return "%s%s (%s)%s" % (postr, self.type, self.tid, synstr)

    @staticmethod
    def from_string(pred):
        # parse off "part_of" specification, if any. This is the remainder
        # of the specification from the first occurrence of part-of.
        part_of_spec = None
        if re.search(r'part[_-]of', pred):
            m = re.match(r'^(.*?)\s*(-+part[_-]of-+>?.*)$', pred)
            assert m, "Error: failed to parse part_of specification: %s" % pred
            pred, part_of_spec = m.groups()

        # parse off synonym specification, if any
        syn_spec = None
        if "synonym" in pred:
            # accept two formats: a square-bracketed string at the end
            # of the string or a dash-separated one at its beginning.
            m = re.match(r'^(.*?)\s*\[([^\[\]]*synonym[^\[\]]*)\]\s*$', pred)
            if m:
                pred, syn_spec = m.groups()
            else:
                m = re.match(r'^\s*-([^-]*synonym[^-]*)-\s*(.*)$', pred)
                assert m, "Error: failed to parse 'synonym' specification on line: %s" % pred
                syn_spec, pred = m.groups()

        # the prediction should now split by tab into strings of the
        # format "TERM (OBOID)" interspersed by "is_a" strings. The
        # last of the strings is the most general predicted type.
        isa_chain = pred.split("\t")

        # sanity
        for i in range(1,len(isa_chain), 2):
            assert isa_chain[i] == "-is_a-", "ERROR: expected \"-is_a-\" in chain, got \"%s\"" % (isa_chain[i])

        # the initially matched term is the first, the "predicted"
        # general type is the last.
        initial_match  = isa_chain[0]
        predicted_type = isa_chain[-1]

        m = re.match(r'^\s*(\S.*?)\s+\((\S+)\)\s*$', initial_match)
        assert m, "Error: failed to parse TERM (ID) format: %s" % initial_match
        match_type, match_id = m.groups()
                
        # parse into term and OBO ID
        m = re.match(r'^\s*(\S.*?)\s+\((\S+)\)\s*$', predicted_type)
        assert m, "Error: failed to parse TERM (ID) format: %s" % predicted_type
        pred_type, pred_id = m.groups()

        # minimal name normalization ...
        pred_type = pred_type.lower()

        # parse the ontology prefix off the initial match ID
        m = re.match(r'^([A-Za-z0-9_-]+).*', match_id)
        assert m, "Error: failed to parse ontology prefix: %s" % pred.rstrip()
        onto = m.group(1)

        return Prediction(pred_type, pred_id, onto, syn_spec, part_of_spec)

def main(argv=None):
    arg = argparser().parse_args(argv[1:])

    # arg coherence
    if arg.matching_only and arg.evaluate_against is None:
        print >> sys.stderr, "Error: -m option requires -E"
        return 1

    # exclude any resource / only include specific?
    excluded_onto = []
    only_include_onto = []
    if arg.exclude is not None:
        if arg.only_include is not None:
            print >> sys.stderr, "Error: -e and -o options are mutually exclusive"
            return 1
        excluded_onto = arg.exclude.split(",")
        # case-insensitive
        excluded_onto = [o.upper() for o in excluded_onto]
    elif arg.only_include is not None:
        only_include_onto = arg.only_include.split(",")
        # case-insensitive
        only_include_onto = [o.upper() for o in only_include_onto]

    targets = None
    target_order = None
    # a closed set of target types specified?
    if arg.targets is not None:
        targets = {}
        target_order = []
        with open(arg.targets) as f:
            for l in f:
                l = l.strip().lower()
                if l in targets:
                    print >> sys.sterr, "Warning: target %s specified multiple times in %s" % (l, arg.targets)
                else:
                    target_order.append(l)
                targets[l] = True

    # "gold" labels to evaluate against defined?
    gold_label = None
    if arg.evaluate_against is not None:
        gold_label = {}
        with open(arg.evaluate_against) as f:
            for ln, l in enumerate(f):
                # assume first non-space character is label and the
                # following space-separated string the term text
                m = re.match(r'^\s*(\S)\s+(\S.*?)\s*$', l)
                assert m, "Failed to parse labeled term in %s line %d: %s" % (arg.evaluate_against, ln+1, l)
                label, term = m.groups()
                assert term not in gold_label, "Error: label for %s defined multiple times in %s" % (term, arg.evaluate_against)
                gold_label[term] = label

    # filter?
    filtered_term = []
    if arg.filter:
        filtered_term = arg.filter.split(',')
        
    total_queries = 0
    complete_misses = 0
    complete_miss_term = {}
    node_misses = 0
    node_miss_term = {}
    with open(arg.file[0]) as f:
        groups = []
        current_group = []
        for l in f:
            m = re.match(r'^\t?(?:Retrieved|Matched) (\d+) strings and (?:retrieved )?(\d+) matches for query:\s*(.*?)\s*$', l)
            if m:
                # end of group
                string_matches, node_matches, query_str = m.groups()
                string_matches, node_matches = int(string_matches), int(node_matches)

                total_queries += 1

                if string_matches == 0:
                    #print >> sys.stderr, "Missed '%s'!" % query_str
                    complete_miss_term[query_str] = True
                    complete_misses += 1
                elif node_matches == 0 or current_group == []:
                    #print >> sys.stderr, "No nodes for '%s'!" % query_str
                    node_miss_term[query_str] = True
                    node_misses += 1
                else:
                    groups.append((query_str, current_group))
                current_group = []

            else:
                # assume normal match. Should have at least three
                # TAB-separated fields: response string, match
                # confidence, and predicted type, which may itself
                # consist of multiple tab-separated strings.
                fields = l.strip().split('\t')
                assert len(fields) >= 3, "Error: expected 3 TAB-separated fields, got %d on line: %s" % (len(fields), l)
                response, confidence, pred = fields[0], fields[1], "\t".join(fields[2:])

                current_group.append((response, Prediction.from_string(pred)))

        # leftovers?
        if current_group != []:
            #groups.append((None, current_group))
            print >> sys.stderr, "Warning: last group not finished; excluding"

        ##
        # filtering

        # filtering by ontology
        filtered_excluded_ontology = 0
        excluded_ontology_term = {}
        if excluded_onto != [] or only_include_onto != []:
            filtered_groups = []
            for query, group in groups:
                filtered_group = []
                for resp, p in group:
                    # upper() for case-insensitive matching
                    if ((excluded_onto != [] and p.onto.upper() not in excluded_onto) or
                        (p.onto.upper() in only_include_onto)):
                        filtered_group.append((resp, p))
                    else:
                        #print >> sys.stderr, "onto-filter: '%s' %s" % (resp, p)
                        pass
                if filtered_group == []:
                    filtered_excluded_ontology += 1
                else:
                    filtered_groups.append((query, filtered_group))
            groups = filtered_groups

        # part_of predictions should normally not be considered. These
        # arise from resources such as BTO that (largely) lack is_a
        # structure and usually lack sufficient specificity to be
        # useful; e.g. "aorta -part_of- animal (BTO:0000042)"
        filtered_part_of_only = 0
        if not arg.include_part_of:
            filtered_groups = []
            for query, group in groups:
                filtered_group = []
                for resp, p in group:
                    if p.part_of_spec is None:
                        filtered_group.append((resp, p))
                    else:
                        #print >> sys.stderr, "part-of-filter: '%s' %s" % (resp, p)
                        pass
                if filtered_group == []:
                    filtered_part_of_only += 1
                else:
                    filtered_groups.append((query, filtered_group))
            groups = filtered_groups

        # By default, synonym match-based predictions are excluded
        # unless they are exact or of a known "safe" type
        filtered_nonexact_synonym_only = 0
        if not arg.include_all_synonyms:
            filtered_groups = []
            for query, group in groups:
                filtered_group = []
                for resp, p in group:
                    if (p.syn_spec is None or 
                        "exact" in p.syn_spec or
                        p.syn_spec in SAFE_SYNONYM_TYPE):
                        filtered_group.append((resp, p))
                    else:
                        #print >> sys.stderr, "synonym-filter: '%s' %s" % (resp, p)
                        pass
                if filtered_group == []:
                    filtered_nonexact_synonym_only += 1
                else:
                    filtered_groups.append((query, filtered_group))
            groups = filtered_groups

        # filtered specific terms
        filtered_term_only = 0
        if arg.filter is not None:
            filtered_groups = []
            for query, group in groups:
                filtered_group = []
                for resp, p in group:
                    if p.type in filtered_term or p.tid in filtered_term:
                        #print >> sys.stderr, "specific-term-filter: '%s' %s" % (resp, p)
                        pass
                    else:
                        filtered_group.append((resp, p))
                if filtered_group == []:
                    #print >> sys.stderr, "Filtered out response to '%s'" % query
                    filtered_term_only += 1
                else:
                    filtered_groups.append((query, filtered_group))
            groups = filtered_groups

        filtered_false_positive = 0
        if arg.evaluate_against:
            # take precision and recall, optionally filter out FP
            filtered_groups = []
            true_positive, false_positive = 0, 0
            true_negative, false_negative = 0, 0
            matched_str = {}
            for query, group in groups:
                if query not in gold_label:
                    print >> sys.stderr, "Warning: no gold label for '%s'; ignoring" % query
                else:
                    matched_str[query] = True
                    # label "a" is for "anatomy" (i.e. TP), any other
                    # considered out.
                    if gold_label[query] == "a":
                        filtered_groups.append((query, group))
                        true_positive += 1
                        if arg.verbose:
                            print "TP:", query
                    else:
                        false_positive += 1
                        if arg.verbose:
                            print "FP:", query

            for query in gold_label:
                if query in matched_str:
                    # already counted
                    continue
                if gold_label[query] == "a":
                    false_negative += 1
                    if arg.verbose:
                        print "FN:", query
                else:
                    true_negative += 1

            if true_positive + false_positive != 0:
                precision = 1.0*true_positive / (true_positive+false_positive)
            else:
                precision = 0.0
            if true_positive + false_negative != 0:
                recall = 1.0*true_positive / (true_positive+false_negative)
            else:
                recall = 0.0
            if precision + recall > 0.0:
                fscore = 2.0*precision*recall / (precision+recall)
            else:
                fscore = 0.0
            print "precision %5.1f%% recall %5.1f%% fscore %5.1f%% (%d TP %d FP %d FN)" % (100.0*precision, 100.0*recall, 100.0*fscore, true_positive, false_positive, false_negative)

            total = true_positive + false_positive + true_negative + false_negative
            if total > 0:
                accuracy = (1.0*true_positive+true_negative) / total
            else:
                accuracy = 0.0
            print "accuracy %.1f%% (%d / %d)" % (100.0*accuracy, true_positive+true_negative, total)

            # remove FP from further processing?
            if arg.matching_only:
                filtered_false_positive = false_positive
                groups = filtered_groups

        print >> sys.stderr, "Analysing %d groups of predictions ..." % len(groups)

        # if there's a closed set of target types, relabel everything
        # else as OTHER_TYPE_STR ("OTHER"). Take and output some stats
        # on most common "other" types
        if targets is not None:
            other_type_count = {}
            target_count, other_count = 0, 0
            for query, group in groups:
                for resp, p in group:
                    if p.type in targets:
                        target_count += 1
                    else:
                        # miss
                        if p.type not in other_type_count:
                            other_type_count[p.type] = 0
                        other_type_count[p.type] += 1
                        p.type = OTHER_TYPE_STR
                        other_count += 1
            print >> sys.stderr, "Found %d instances of target types, %d of other" % (target_count, other_count)

            ordered_other_types = other_type_count.keys()
            ordered_other_types.sort(lambda a,b: cmp(other_type_count[b],other_type_count[a]))
            print >> sys.stderr, "Top other types:", ", ".join("%s:%d" % (o, other_type_count[o]) for o in ordered_other_types[:5])

        # determine how many types there are altogether and their
        # occurrence counts.
        type_count = {}
        for query, group in groups:
            for resp, p in group:
                if p.type not in type_count:
                    type_count[p.type] = 0
                type_count[p.type] += 1

        # output type counts in decreasing order for number of
        # occurrences
        sorted_types = type_count.keys()
        sorted_types.sort(lambda a,b: cmp(type_count[b],type_count[a]))

        # ... except if targets are defined, in which case order in
        # the given file is used, with OTHER_TYPE_STR ("OTHER") last.
        if target_order is not None:
            target_order.append(OTHER_TYPE_STR)
            sorted_types.sort(lambda a,b: cmp(target_order.index(a),target_order.index(b)))

#         print "Types:"
#         for t in sorted_types:
#             print "%d\t%s" % (type_count[t], t)

        # do a confusion matrix, storing for each of the predicted
        # types the number of times it co-occurs in a group with
        # another.
        matrix = {}
        for t1 in sorted_types:
            matrix[t1] = {}
            for t2 in sorted_types:
                matrix[t1][t2] = 0

        for query, group in groups:
            for i1 in range(len(group)):
                for i2 in range(len(group)):
                    if i1 == i2:
                        continue
                    r1, p1 = group[i1]
                    r2, p2 = group[i2]

                    matrix[p1.type][p2.type] += 1

        if arg.confusion_matrix:
            # print confusion matrix in TAB-separated CSV
            for i, t1 in enumerate(sorted_types):
                print "%d\t%30s\t" % (i, t1) + "\t".join([str(matrix[t1][t2]) for t2 in sorted_types])

        if arg.output_types:
            # print types for each response
            for query, group in groups:                
                print query, "\t", 
                ontos_for_type = {}
                for r, p in group:
                    if p.type not in ontos_for_type:
                        ontos_for_type[p.type] = []
                    ontos_for_type[p.type].append(p.onto)
                print ", ".join(["%s (%s)" % (t, ",".join(ontos_for_type[t])) for t in ontos_for_type])

        # take summary statistics: what's the average number of
        # responses and the average number of different classes
        # assigned per query
        total_responses = 0
        total_ontologies = 0
        total_types = 0
        for query, group in groups:
            total_responses += len(group)
            total_types += len(set([p.type for r,p in group]))
            total_ontologies += len(set([p.onto for r,p in group]))
        if len(groups) == 0:
            print >> sys.stderr, "Nothing evaluated!"
        else:
            avg_resp  = 1.*total_responses/len(groups)
            avg_onto  = 1.*total_ontologies/len(groups)
            avg_types = 1.*total_types/len(groups)
            print >> sys.stderr, "Average number of responses:    %f" % avg_resp
            print >> sys.stderr, "Average number of ontologies:   %f" % avg_onto
            print >> sys.stderr, "Average number of unique types: %f" % avg_types
            print >> sys.stderr, "Average response/type ratio:    %f" % (avg_resp/avg_types)
            print >> sys.stderr, "Average ontology/type ratio:    %f" % (avg_onto/avg_types)

        # sanity
        assert total_queries == complete_misses + node_misses + filtered_excluded_ontology + filtered_part_of_only + filtered_nonexact_synonym_only + filtered_term_only + filtered_false_positive+ len(groups), "INTERNAL ERROR"

        # general stats
        print >> sys.stderr, "Processes %d queries:" % total_queries
        print >> sys.stderr, "\t%d matched nothing" % complete_misses
        print >> sys.stderr, "\t%d matched only non-relevant terms" % node_misses
        print >> sys.stderr, "\t%d matched only excluded ontologies" % filtered_excluded_ontology
        print >> sys.stderr, "\t%d matched only filtered terms" % filtered_term_only
        print >> sys.stderr, "\t%d filtered for only part-of responses" % filtered_part_of_only
        print >> sys.stderr, "\t%d filtered for only non-exact synonym responses" % filtered_nonexact_synonym_only
        print >> sys.stderr, "\t%d filtered as not matching gold" % filtered_false_positive
        print >> sys.stderr, "\t%d evaluated" % len(groups)


if __name__ == "__main__":
    sys.exit(main(sys.argv))
